==> ?obj=sdir%2F1.21%2Fsdir-source.zip: sdir-1.21/utils.c <==
/*
 * UTILS.C
 *
 * Colorful, sorted and optionally rich directory enumeration
 * for Windows.
 *
 * This module contains helper functions that can be used in a variety of
 * situations.
 *
 * Copyright (c) 2014 Malcolm J. Smith
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "sdir.h"

//
//  Sorting support
//

DWORD
SdirCompareLargeInt (
    __in PULARGE_INTEGER Left,
    __in PULARGE_INTEGER Right
    )
{
    if (Left->HighPart < Right->HighPart) {
        return SDIR_LESS_THAN;
    } else if (Left->HighPart > Right->HighPart) {
        return SDIR_GREATER_THAN;
    }

    if (Left->LowPart < Right->LowPart) {
        return SDIR_LESS_THAN;
    } else if (Left->LowPart > Right->LowPart) {
        return SDIR_GREATER_THAN;
    }

    return SDIR_EQUAL;
}

DWORD
SdirCompareString (
    __in LPCTSTR Left,
    __in LPCTSTR Right
    )
{
    int ret = _tcsicmp(Left, Right);

    if (ret < 0) {
        return SDIR_LESS_THAN;
    } else if (ret == 0) {
        return SDIR_EQUAL;
    } else {
        return SDIR_GREATER_THAN;
    }
}

DWORD
SdirCompareDate (
    __in LPSYSTEMTIME Left,
    __in LPSYSTEMTIME Right
    )
{
    if (Left->wYear < Right->wYear) {
        return SDIR_LESS_THAN;
    } else if (Left->wYear > Right->wYear) {
        return SDIR_GREATER_THAN;
    }

    if (Left->wMonth < Right->wMonth) {
        return SDIR_LESS_THAN;
    } else if (Left->wMonth > Right->wMonth) {
        return SDIR_GREATER_THAN;
    }

    if (Left->wDay < Right->wDay) {
        return SDIR_LESS_THAN;
    } else if (Left->wDay > Right->wDay) {
        return SDIR_GREATER_THAN;
    }

    return SDIR_EQUAL;
}

DWORD
SdirCompareTime (
    __in LPSYSTEMTIME Left,
    __in LPSYSTEMTIME Right
    )
{
    if (Left->wHour < Right->wHour) {
        return SDIR_LESS_THAN;
    } else if (Left->wHour > Right->wHour) {
        return SDIR_GREATER_THAN;
    }

    if (Left->wMinute < Right->wMinute) {
        return SDIR_LESS_THAN;
    } else if (Left->wMinute > Right->wMinute) {
        return SDIR_GREATER_THAN;
    }

    if (Left->wSecond < Right->wSecond) {
        return SDIR_LESS_THAN;
    } else if (Left->wSecond > Right->wSecond) {
        return SDIR_GREATER_THAN;
    }

    if (Left->wMilliseconds < Right->wMilliseconds) {
        return SDIR_LESS_THAN;
    } else if (Left->wMilliseconds > Right->wMilliseconds) {
        return SDIR_GREATER_THAN;
    }

    return SDIR_EQUAL;
}

//
//  These are general helper routines to assist in converting strings into
//  something we can use directly.
//

DWORD
SdirStringToNum32(
    __in TCHAR * str,
    __out_opt TCHAR ** endptr
    )
{
    DWORD ret = 0;
    if (str[0] == '0' && str[1] == 'x') {
        str += 2;
        while ((*str >= '0' && *str <= '9') || (*str >= 'a' && *str <= 'f') || (*str >= 'A' && *str <= 'F')) {
            ret *= 16;
            if (*str >= '0' && *str <= '9') {
                ret += *str - '0';
            } else if (*str >= 'a' && *str <= 'f') {
                ret += *str - 'a' + 10;
            } else if (*str >= 'A' && *str <= 'F') {
                ret += *str - 'A' + 10;
            }
            str++;
        }
    } else {
        while (*str >= '0' && *str <= '9') {
            ret *= 10;
            ret += *str - '0';
            str++;
        }
    }
    if (endptr) {
        *endptr = str;
    }
    return ret;
}

LARGE_INTEGER
SdirStringToNum64(
    __in TCHAR * str,
    __out_opt TCHAR ** endptr
    )
{
    LARGE_INTEGER ret = {0};
    if (str[0] == '0' && str[1] == 'x') {
        str += 2;
    }
    while ((*str >= '0' && *str <= '9') || (*str >= 'a' && *str <= 'f') || (*str >= 'A' && *str <= 'F')) {
        ret.HighPart = (ret.HighPart<<4) + ((ret.LowPart & 0xf0000000)>>28);
        ret.LowPart *= 16;
        if (*str >= '0' && *str <= '9') {
            ret.LowPart += *str - '0';
        } else if (*str >= 'a' && *str <= 'f') {
            ret.LowPart += *str - 'a' + 10;
        } else if (*str >= 'A' && *str <= 'F') {
            ret.LowPart += *str - 'A' + 10;
        }
        str++;
    }
    if (endptr) {
        *endptr = str;
    }
    return ret;
}

LARGE_INTEGER
SdirStringToFileSize(
    __in TCHAR * str
    )
{
    TCHAR Suffixes[] = {'b', 'k', 'm', 'g', 't'};
    DWORD SuffixLevel = 0;
    LARGE_INTEGER FileSize;
    DWORD i;
    TCHAR * endptr;

    FileSize.HighPart = 0;
    FileSize.LowPart = SdirStringToNum32(str, &endptr);

    for (i = 0; i < sizeof(Suffixes)/sizeof(Suffixes[0]); i++) {
        if (*endptr == Suffixes[i] || *endptr == (Suffixes[i] + 'A' - 'a')) {
            SuffixLevel = i;
            break;
        }
    }

    while(SuffixLevel) {

        FileSize.HighPart = (FileSize.HighPart << 10) + (FileSize.LowPart >> 22);
        FileSize.LowPart = (FileSize.LowPart << 10);

        SuffixLevel--;
    }

    return FileSize;
}

BOOL
SdirStringToDate(
    __in TCHAR * str,
    __out PSYSTEMTIME date
    )
{
    TCHAR * endptr = NULL;
    date->wYear = (WORD)SdirStringToNum32(str, &endptr);
    if (date->wYear < 100) {
        date->wYear += 2000;
    }
    if (*endptr == '/') {
        date->wMonth = (WORD)SdirStringToNum32(endptr + 1, &endptr);
        if (*endptr == '/') {
            date->wDay = (WORD)SdirStringToNum32(endptr + 1, &endptr);
        }
    }

    return TRUE;
}

BOOL
SdirStringToTime(
    __in TCHAR * str,
    __out PSYSTEMTIME date
    )
{
    TCHAR * endptr = NULL;
    date->wHour = (WORD)SdirStringToNum32(str, &endptr);
    if (*endptr == ':') {
        date->wMinute = (WORD)SdirStringToNum32(endptr + 1, &endptr);
        if (*endptr == ':') {
            date->wSecond = (WORD)SdirStringToNum32(endptr + 1, &endptr);
        }
    }

    return TRUE;
}

//
//  Formatting support.  Generic routines used across different types of data.
//

ULONG
SdirDisplayGenericSize(
    __out_ecount(6) PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PLARGE_INTEGER GenericSize
    )
{
    TCHAR Suffixes[] = {'b', 'k', 'm', 'g', 't', '?'};
    DWORD SuffixLevel = 0;
    TCHAR StrSize[10];
        
    if (Buffer) {
        LARGE_INTEGER Size = *GenericSize;
        LARGE_INTEGER OldSize = Size;

        while (Size.HighPart != 0 || Size.LowPart > 9999) {
            SuffixLevel++;
            OldSize = Size;

            //
            //  Conceptually we want to divide by 1024.  We do this
            //  in two 32-bit shifts combining back the high bits
            //  so we don't need full compiler support.
            //

            Size.LowPart = (Size.LowPart >> 10) + ((Size.HighPart & 0x3ff) << 22);
            Size.HighPart = Size.HighPart >> 10;
        }

        if (SuffixLevel >= sizeof(Suffixes)/sizeof(TCHAR)) {
            SuffixLevel = sizeof(Suffixes)/sizeof(TCHAR) - 1;
        }
            
        if (Size.LowPart < 100 && SuffixLevel > 0) {
            OldSize.LowPart = (OldSize.LowPart % 1024) * 10 / 1024;
            _stprintf_s(StrSize,
                        sizeof(StrSize)/sizeof(StrSize[0]),
                        _T(" %2i.%1i%c"),
                        Size.LowPart,
                        OldSize.LowPart,
                        Suffixes[SuffixLevel]);
        } else {
            _stprintf_s(StrSize,
                        sizeof(StrSize)/sizeof(StrSize[0]),
                        _T(" %4i%c"),
                        Size.LowPart,
                        Suffixes[SuffixLevel]);
        }
            
        SdirPasteStr(Buffer, StrSize, Attributes, 6);
    }
    return 6;
}

ULONG
SdirDisplayHex64 (
    __out_ecount(18) PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PLARGE_INTEGER Hex
    )
{
    TCHAR Str[19];

    if (Buffer) {
        _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" %08x`%08x"), Hex->HighPart, Hex->LowPart);
        SdirPasteStr(Buffer, Str, SDIR_FILE_TIME_COLOR, 18);
    }

    return 18;
}

ULONG
SdirDisplayHex32 (
    __out_ecount(9) PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in DWORD Hex
    )
{
    TCHAR Str[10];

    if (Buffer) {
        _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" %08x"), Hex);
        SdirPasteStr(Buffer, Str, SDIR_FILE_TIME_COLOR, 9);
    }
    return 9;
}

ULONG
SdirDisplayFileDate (
    __out_ecount(11) PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in SYSTEMTIME * Time
    )
{
    TCHAR StrDate[16];

    if (Buffer) {
        _stprintf_s(StrDate,
                    sizeof(StrDate)/sizeof(StrDate[0]),
                    _T(" %04i/%02i/%02i"),
                    Time->wYear,
                    Time->wMonth,
                    Time->wDay);

        SdirPasteStr(Buffer, StrDate, SDIR_FILE_TIME_COLOR, 11);
    }
    return 11;
}

ULONG
SdirDisplayFileTime (
    __out_ecount(9) PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in SYSTEMTIME * Time
    )
{
    TCHAR StrTime[10];

    if (Buffer) {
        _stprintf_s(StrTime,
                    sizeof(StrTime)/sizeof(StrTime[0]),
                    _T(" %02i:%02i:%02i"),
                    Time->wHour,
                    Time->wMinute,
                    Time->wSecond);

        SdirPasteStr(Buffer, StrTime, SDIR_FILE_TIME_COLOR, 9);
    }
    return 9;
}

// vim:sw=4:ts=4:et:

==> ?obj=sdir%2F1.25%2Fsdir-source.zip: sdir-1.25/utils.c <==
/*
 * UTILS.C
 *
 * Colorful, sorted and optionally rich directory enumeration
 * for Windows.
 *
 * This module contains helper functions that can be used in a variety of
 * situations.
 *
 * Copyright (c) 2014 Malcolm J. Smith
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "sdir.h"

//
//  Sorting support
//

DWORD
SdirCompareLargeInt (
    __in PULARGE_INTEGER Left,
    __in PULARGE_INTEGER Right
    )
{
    if (Left->HighPart < Right->HighPart) {
        return SDIR_LESS_THAN;
    } else if (Left->HighPart > Right->HighPart) {
        return SDIR_GREATER_THAN;
    }

    if (Left->LowPart < Right->LowPart) {
        return SDIR_LESS_THAN;
    } else if (Left->LowPart > Right->LowPart) {
        return SDIR_GREATER_THAN;
    }

    return SDIR_EQUAL;
}

DWORD
SdirCompareString (
    __in LPCTSTR Left,
    __in LPCTSTR Right
    )
{
    int ret = _tcsicmp(Left, Right);

    if (ret < 0) {
        return SDIR_LESS_THAN;
    } else if (ret == 0) {
        return SDIR_EQUAL;
    } else {
        return SDIR_GREATER_THAN;
    }
}

DWORD
SdirCompareDate (
    __in LPSYSTEMTIME Left,
    __in LPSYSTEMTIME Right
    )
{
    if (Left->wYear < Right->wYear) {
        return SDIR_LESS_THAN;
    } else if (Left->wYear > Right->wYear) {
        return SDIR_GREATER_THAN;
    }

    if (Left->wMonth < Right->wMonth) {
        return SDIR_LESS_THAN;
    } else if (Left->wMonth > Right->wMonth) {
        return SDIR_GREATER_THAN;
    }

    if (Left->wDay < Right->wDay) {
        return SDIR_LESS_THAN;
    } else if (Left->wDay > Right->wDay) {
        return SDIR_GREATER_THAN;
    }

    return SDIR_EQUAL;
}

DWORD
SdirCompareTime (
    __in LPSYSTEMTIME Left,
    __in LPSYSTEMTIME Right
    )
{
    if (Left->wHour < Right->wHour) {
        return SDIR_LESS_THAN;
    } else if (Left->wHour > Right->wHour) {
        return SDIR_GREATER_THAN;
    }

    if (Left->wMinute < Right->wMinute) {
        return SDIR_LESS_THAN;
    } else if (Left->wMinute > Right->wMinute) {
        return SDIR_GREATER_THAN;
    }

    if (Left->wSecond < Right->wSecond) {
        return SDIR_LESS_THAN;
    } else if (Left->wSecond > Right->wSecond) {
        return SDIR_GREATER_THAN;
    }

    if (Left->wMilliseconds < Right->wMilliseconds) {
        return SDIR_LESS_THAN;
    } else if (Left->wMilliseconds > Right->wMilliseconds) {
        return SDIR_GREATER_THAN;
    }

    return SDIR_EQUAL;
}

//
//  These are general helper routines to assist in converting strings into
//  something we can use directly.
//

DWORD
SdirStringToNum32(
    __in TCHAR * str,
    __out_opt TCHAR ** endptr
    )
{
    DWORD ret = 0;
    if (str[0] == '0' && str[1] == 'x') {
        str += 2;
        while ((*str >= '0' && *str <= '9') || (*str >= 'a' && *str <= 'f') || (*str >= 'A' && *str <= 'F')) {
            ret *= 16;
            if (*str >= '0' && *str <= '9') {
                ret += *str - '0';
            } else if (*str >= 'a' && *str <= 'f') {
                ret += *str - 'a' + 10;
            } else if (*str >= 'A' && *str <= 'F') {
                ret += *str - 'A' + 10;
            }
            str++;
        }
    } else {
        while (*str >= '0' && *str <= '9') {
            ret *= 10;
            ret += *str - '0';
            str++;
        }
    }
    if (endptr) {
        *endptr = str;
    }
    return ret;
}

LARGE_INTEGER
SdirStringToNum64(
    __in TCHAR * str,
    __out_opt TCHAR ** endptr
    )
{
    LARGE_INTEGER ret = {0};
    if (str[0] == '0' && str[1] == 'x') {
        str += 2;
    }
    while ((*str >= '0' && *str <= '9') || (*str >= 'a' && *str <= 'f') || (*str >= 'A' && *str <= 'F')) {
        ret.HighPart = (ret.HighPart<<4) + ((ret.LowPart & 0xf0000000)>>28);
        ret.LowPart *= 16;
        if (*str >= '0' && *str <= '9') {
            ret.LowPart += *str - '0';
        } else if (*str >= 'a' && *str <= 'f') {
            ret.LowPart += *str - 'a' + 10;
        } else if (*str >= 'A' && *str <= 'F') {
            ret.LowPart += *str - 'A' + 10;
        }
        str++;
    }
    if (endptr) {
        *endptr = str;
    }
    return ret;
}

LARGE_INTEGER
SdirStringToFileSize(
    __in TCHAR * str
    )
{
    TCHAR Suffixes[] = {'b', 'k', 'm', 'g', 't'};
    DWORD SuffixLevel = 0;
    LARGE_INTEGER FileSize;
    DWORD i;
    TCHAR * endptr;

    FileSize.HighPart = 0;
    FileSize.LowPart = SdirStringToNum32(str, &endptr);

    for (i = 0; i < sizeof(Suffixes)/sizeof(Suffixes[0]); i++) {
        if (*endptr == Suffixes[i] || *endptr == (Suffixes[i] + 'A' - 'a')) {
            SuffixLevel = i;
            break;
        }
    }

    while(SuffixLevel) {

        FileSize.HighPart = (FileSize.HighPart << 10) + (FileSize.LowPart >> 22);
        FileSize.LowPart = (FileSize.LowPart << 10);

        SuffixLevel--;
    }

    return FileSize;
}

BOOL
SdirStringToDate(
    __in TCHAR * str,
    __out PSYSTEMTIME date
    )
{
    TCHAR * endptr = NULL;
    date->wYear = (WORD)SdirStringToNum32(str, &endptr);
    if (date->wYear < 100) {
        date->wYear += 2000;
    }
    if (*endptr == '/') {
        date->wMonth = (WORD)SdirStringToNum32(endptr + 1, &endptr);
        if (*endptr == '/') {
            date->wDay = (WORD)SdirStringToNum32(endptr + 1, &endptr);
        }
    }

    return TRUE;
}

BOOL
SdirStringToTime(
    __in TCHAR * str,
    __out PSYSTEMTIME date
    )
{
    TCHAR * endptr = NULL;
    date->wHour = (WORD)SdirStringToNum32(str, &endptr);
    if (*endptr == ':') {
        date->wMinute = (WORD)SdirStringToNum32(endptr + 1, &endptr);
        if (*endptr == ':') {
            date->wSecond = (WORD)SdirStringToNum32(endptr + 1, &endptr);
        }
    }

    return TRUE;
}

//
//  Formatting support.  Generic routines used across different types of data.
//

ULONG
SdirDisplayGenericSize(
    __out_ecount(6) PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PLARGE_INTEGER GenericSize
    )
{
    TCHAR Suffixes[] = {'b', 'k', 'm', 'g', 't', '?'};
    DWORD SuffixLevel = 0;
    TCHAR StrSize[10];
        
    if (Buffer) {
        LARGE_INTEGER Size = *GenericSize;
        LARGE_INTEGER OldSize = Size;

        while (Size.HighPart != 0 || Size.LowPart > 9999) {
            SuffixLevel++;
            OldSize = Size;

            //
            //  Conceptually we want to divide by 1024.  We do this
            //  in two 32-bit shifts combining back the high bits
            //  so we don't need full compiler support.
            //

            Size.LowPart = (Size.LowPart >> 10) + ((Size.HighPart & 0x3ff) << 22);
            Size.HighPart = Size.HighPart >> 10;
        }

        if (SuffixLevel >= sizeof(Suffixes)/sizeof(TCHAR)) {
            SuffixLevel = sizeof(Suffixes)/sizeof(TCHAR) - 1;
        }
            
        if (Size.LowPart < 100 && SuffixLevel > 0) {
            OldSize.LowPart = (OldSize.LowPart % 1024) * 10 / 1024;
            _stprintf_s(StrSize,
                        sizeof(StrSize)/sizeof(StrSize[0]),
                        _T(" %2i.%1i%c"),
                        Size.LowPart,
                        OldSize.LowPart,
                        Suffixes[SuffixLevel]);
        } else {
            _stprintf_s(StrSize,
                        sizeof(StrSize)/sizeof(StrSize[0]),
                        _T(" %4i%c"),
                        Size.LowPart,
                        Suffixes[SuffixLevel]);
        }
            
        SdirPasteStr(Buffer, StrSize, Attributes, 6);
    }
    return 6;
}

ULONG
SdirDisplayHex64 (
    __out_ecount(18) PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PLARGE_INTEGER Hex
    )
{
    TCHAR Str[19];

    if (Buffer) {
        _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" %08x`%08x"), Hex->HighPart, Hex->LowPart);
        SdirPasteStr(Buffer, Str, Attributes, 18);
    }

    return 18;
}

ULONG
SdirDisplayHex32 (
    __out_ecount(9) PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in DWORD Hex
    )
{
    TCHAR Str[10];

    if (Buffer) {
        _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" %08x"), Hex);
        SdirPasteStr(Buffer, Str, Attributes, 9);
    }
    return 9;
}

ULONG
SdirDisplayFileDate (
    __out_ecount(11) PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in SYSTEMTIME * Time
    )
{
    TCHAR StrDate[16];

    if (Buffer) {
        _stprintf_s(StrDate,
                    sizeof(StrDate)/sizeof(StrDate[0]),
                    _T(" %04i/%02i/%02i"),
                    Time->wYear,
                    Time->wMonth,
                    Time->wDay);

        SdirPasteStr(Buffer, StrDate, Attributes, 11);
    }
    return 11;
}

ULONG
SdirDisplayFileTime (
    __out_ecount(9) PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in SYSTEMTIME * Time
    )
{
    TCHAR StrTime[10];

    if (Buffer) {
        _stprintf_s(StrTime,
                    sizeof(StrTime)/sizeof(StrTime[0]),
                    _T(" %02i:%02i:%02i"),
                    Time->wHour,
                    Time->wMinute,
                    Time->wSecond);

        SdirPasteStr(Buffer, StrTime, Attributes, 9);
    }
    return 9;
}

// vim:sw=4:ts=4:et:
